"""Common tools to supplement/assist the standard logging package."""

import argparse
import logging
from typing import List, Optional, Union

from typing_extensions import Literal  # will redirect to Typing for 3.8+

LoggerLevel = Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]


def log_argparse_args(
    args: argparse.Namespace,
    logger: Optional[Union[str, logging.Logger]] = None,
    level: LoggerLevel = "WARNING",
) -> argparse.Namespace:
    """Log the argparse args and their values at the given level.

    Return the args (Namespace) unchanged.

    Example:
        2022-05-13 22:37:21 fv-az136-643 my-logs[61] WARNING in_file: in_msg.pkl
        2022-05-13 22:37:21 fv-az136-643 my-logs[61] WARNING out_file: out_msg.pkl
        2022-05-13 22:37:21 fv-az136-643 my-logs[61] WARNING log: DEBUG
        2022-05-13 22:37:21 fv-az136-643 my-logs[61] WARNING log_third_party: WARNING
    """
    if not logger:
        _logger = logging.getLogger()
    elif isinstance(logger, logging.Logger):
        _logger = logger
    else:
        _logger = logging.getLogger(logger)

    if level not in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]:
        raise ValueError(f"Invalid logging level: {level}")
    log = getattr(_logger, level.lower())  # ..., info, warning, critical, ...
    for arg, val in vars(args).items():
        log(f"{arg}: {val}")

    return args


def set_level(
    level: str,
    first_party_loggers: Optional[List[Union[str, logging.Logger]]] = None,
    third_party_level: LoggerLevel = "WARNING",
    use_coloredlogs: bool = False,
) -> None:
    """Set the level of the root logger, first-party loggers, and third-party loggers.

    The root logger and first-party logger(s) are set to the same level
    (`level`). The third-party loggers are non-root and non-first-party
    loggers that are defined at the time of invocation. If a logger is
    created after this function call, then its level defaults to its
    parent (that's the root logger for non-child loggers).

    Passing `use_coloredlogs=True` will import and use the `coloredlogs`
    package. This will set the logger format and use colored text.
    """
    if not first_party_loggers:
        first_party_loggers = []

    # root
    if use_coloredlogs:
        try:
            import coloredlogs  # type: ignore[import]  # pylint: disable=import-outside-toplevel

            coloredlogs.install(level=level)  # root
        except ImportError:
            logging.getLogger("wipac_dev_tools.logging_tools").warning(
                "set_level()'s `use_coloredlogs` was set to `True`, "
                "but coloredlogs is not installed. Proceeding with only logging package."
            )
            logging.getLogger().setLevel(level)
    else:
        logging.getLogger().setLevel(level)

    # first-party
    for log in first_party_loggers:
        if isinstance(log, logging.Logger):
            log.setLevel(level)
        else:  # str
            logging.getLogger(log).setLevel(level)

    # third-party
    for log_name in logging.root.manager.loggerDict:
        if log_name in first_party_loggers:
            continue
        if logging.getLogger(log_name) in first_party_loggers:
            continue
        logging.getLogger(log_name).setLevel(third_party_level)



"""The useful `strtobool()`, originally from the deprecated `distutils.util`."""


def strtobool(string: str) -> bool:
    """Smart-cast a string to a bool using common-sense interpretations.

    Unlike the since deprecated `distutils.util.strtobool`, this
    returns an actual bool.

    True: 'y', 'yes', 't', 'true', 'on', '1'
    False: 'n', 'no', 'f', 'false', 'off', '0'

    Raises:
        ValueError: if the string does not match any of the about
    """
    if string.lower() in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif string.lower() in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"Invalid truth value: {string}")



"""Module to support parsing environment variables."""


import os
import re
import sys
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from .strtobool import strtobool

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final  # type: ignore[misc]

# IMPORTS for PYTHON 3.7+
if sys.version_info >= (3, 7):
    import dataclasses
    from typing import _SpecialForm

    try:
        from typing import _GenericAlias as GenericAlias  # type: ignore[attr-defined]
    except ImportError:
        from typing import GenericAlias  # type: ignore[attr-defined]


RetVal = Union[str, int, float, bool]
OptionalDict = Mapping[str, Optional[RetVal]]
KeySpec = Union[str, Sequence[str], OptionalDict]


def _typecast(source: str, type_: type) -> RetVal:
    if type_ == bool:
        return bool(strtobool(source.lower()))
    elif type_ == int:
        return int(source)
    elif type_ == float:
        return float(source)
    else:
        return source


def from_environment(keys: KeySpec) -> Dict[str, RetVal]:
    """Obtain configuration values from the OS environment.

    Parsing Details:
    Types are inferred from the default values, and casted as such:
    `bool`: *(case-insensitive)*:
        - `True`  => ("y", "yes", "t", "true", "on", or "1")
        - `False` => ("n", "no", "f", "false", "off", or "0")
        - `Error` => any other string
    `int`: normal cast (`int(str)`)
    `float`: normal cast (`float(str)`)
    `other`: no change (`str`)

    Arguments:
        keys - Specify the configuration values to obtain.

               This can be a string, specifying a single key, such as:

                   config_dict = from_environment("LANGUAGE")

               This can be a list of strings, specifying multiple keys,
               such as:

                   config_dict = from_environment(["HOME", "LANGUAGE"])

               This can be a dictionary that provides some default values,
               and will accept overrides from the environment:

                   default_config = {
                       "HOST": "localhost",
                       "PORT": 8080,
                       "REQUIRED_FROM_ENVIRONMENT": None
                   }
                   config_dict = from_environment(default_config)

               Note in this case that if 'HOST' or 'PORT' were defined in the
               environment, those values would be returned in config_dict. If
               the values were not defined in the environment, the default values
               from default_config would be returned in config_dict.

               Also note, that if 'REQUIRED_FROM_ENVIRONMENT' is not defined,
               an OSError will be raised. The sentinel value of None indicates
               that the configuration parameter MUST be sourced from the
               environment.

    Returns:
        a dictionary mapping configuration keys to configuration values

    Raises:
        OSError - If a configuration value is requested and no default
                  value is provided (via a dict), to indicate that the
                  component's configuration is incomplete due to missing
                  data from the OS.
        ValueError - If a type-indicated value is not a legal value
    """
    if isinstance(keys, str):
        keys = {keys: None}
    elif isinstance(keys, list):
        keys = dict.fromkeys(keys, None)
    elif not isinstance(keys, dict):
        raise TypeError("keys: Expected string, list or dict")

    config = keys.copy()

    for key in config:
        # grab & cast key-value
        if key in os.environ:
            try:
                config[key] = _typecast(os.environ[key], type(config[key]))
            except ValueError:
                raise ValueError(  # pylint: disable=raise-missing-from
                    f"'{type(config[key])}'-indicated value is not a legal value: "
                    f"key='{key}' value='{config[key]}'"
                )
        # missing key
        elif config[key] is None:
            raise OSError(f"Missing environment variable '{key}'")

    return cast(Dict[str, RetVal], config)


def _typecast_for_dataclass(
    env_val: str,
    typ: type,
    arg_typs: Optional[Tuple[type, ...]],
    collection_sep: Optional[str],
    dict_kv_joiner: str,
) -> Any:
    """Collect the typecast value"""
    if typ == list:
        _list = env_val.split(collection_sep)
        if arg_typs:
            return [arg_typs[0](x) for x in _list]
        return _list

    elif typ == dict:
        _dict = {
            x.split(dict_kv_joiner)[0]: x.split(dict_kv_joiner)[1]
            for x in env_val.split(collection_sep)
        }
        if arg_typs:
            return {arg_typs[0](k): arg_typs[1](v) for k, v in _dict.items()}
        return _dict

    elif typ == set:
        _set = set(env_val.split(collection_sep))
        if arg_typs:
            return {arg_typs[0](x) for x in _set}
        return _set

    elif typ == frozenset:
        _frozenset = frozenset(env_val.split(collection_sep))
        if arg_typs:
            return {arg_typs[0](x) for x in _frozenset}
        return _frozenset

    elif typ == bool:
        return strtobool(env_val)

    else:
        return typ(env_val)


T = TypeVar("T")


def from_environment_as_dataclass(
    dclass: Type[T],
    collection_sep: Optional[str] = None,
    dict_kv_joiner: str = "=",
) -> T:
    """Obtain configuration values from the OS environment formatted in a dataclass.

    Environment variables are matched to a dataclass field's name. The
    matching environment string is cast using the dataclass field's type
    (there are some special cases for built-in types, see below). Then,
    the values are used to create a dataclass instance. All normal
    dataclass init-behavior is expected, like required fields
    (positional arguments), optional fields with defaults, default
    factories, post-init processing, etc.

    If a field's type is a bool, `wipac_dev_tools.strtobool` is applied.

    If a field's type is a `list`, `dict`, `set`, `frozenset`, or
    an analogous type alias from the 'typing' module, then a conversion
    is made (see `collection_sep` and `dict_kv_joiner`). Sub-types
    are cast if using a typing-module type alias. The typing-module's
    alias types must resolve to `type` within 1 nesting (eg: List[bool]
    and Dict[int, float] are okay; List[Dict[int, float]] is not), or
    2 if using 'Final' or 'Optional' (ex: Final[Dict[int, float]]).

    If a field's type is a class that accepts 1 argument, it is
    instantiated as such.

    Arguments:
        dclass - a (non-instantiated) dataclass, aka a type
        collection_sep - the delimiter to split collections on ("1 2 5")
        dict_kv_joiner - the delimiter that joins key-value pairs ("a=1 b=2 c=1")

    Returns:
        a dataclass instance mapping configuration keys to configuration values

    Example:
        env:
            FPATH=/home/example/path
            PORT=9999
            HOST=localhost
            MSGS_PER_CLIENTS=alpha=0 beta=55 delta=3
            USE_EVEN=22
            RETRIES=3

        python:
            @dataclasses.dataclass(frozen=True)
            class Config:
                FPATH: pathlib.Path
                PORT: int
                HOST: str
                MSGS_PER_CLIENTS: Dict[str, int]
                USE_EVEN: EvenState
                RETRIES: Optional[int] = None
                TIMEOUT: int = 30

                def __post_init__(self) -> None:
                    if self.PORT <= 0:
                        raise ValueError("'PORT' is non-positive")

            class EvenState:
                def __init__(self, arg: str):
                    self.is_even = not bool(int(arg) % 2)  # 1%2 -> 1 -> T -> F
                def __repr__(self) -> str:
                    return f"EvenState(is_even={self.is_even})"

            config = from_environment_as_dataclass(Config)
            print(config)

        stdout:
            Config(
                FPATH=PosixPath('/home/example/path'),
                PORT=9999,
                HOST='localhost',
                MSGS_PER_CLIENTS={'alpha': 0, 'beta': 55, 'delta': 3},
                USE_EVEN=EvenState(is_even=True),
                RETRIES=3,
                TIMEOUT=30)


    Raises:
        OSError - If a configuration value is requested and no default
                  value is provided, to indicate that the component's
                  configuration is incomplete due to missing data from
                  the OS.
        ValueError - If an indicated value is not a legal value
        TypeError - If an argument or indicated value is not a legal type
    """

    if sys.version_info >= (3, 7):
        return _from_environment_as_dataclass(dclass, collection_sep, dict_kv_joiner)
    else:
        raise NotImplementedError(
            "Sorry, from_environment_as_dataclass() is only available for 3.7+"
        )


def _from_environment_as_dataclass(
    dclass: Type[T],
    collection_sep: Optional[str],
    dict_kv_joiner: str,
) -> T:

    # check args
    if (
        (dict_kv_joiner == collection_sep)
        or (not collection_sep and " " in dict_kv_joiner)  # collection_sep=None is \s+
        or (collection_sep and collection_sep in dict_kv_joiner)
    ):
        raise RuntimeError(
            r"'collection_sep' ('None'='\s+') cannot overlap with 'dict_kv_joiner': "
            f"'{collection_sep}' & '{dict_kv_joiner}'"
        )

    # type-check dclass
    if not (dataclasses.is_dataclass(dclass) and isinstance(dclass, type)):
        raise TypeError(f"Expected (non-instantiated) dataclass: 'dclass' ({dclass})")

    # some helper functions
    def _is_optional(typ: GenericAlias) -> bool:
        # Optional[int] *is* typing.Union[int, NoneType]
        return (
            typ.__origin__ == Union
            and len(typ.__args__) == 2
            and typ.__args__[-1] == type(None)  # noqa: E721
        )

    def _is_final(typ: GenericAlias) -> bool:
        return bool(typ.__origin__ == Final)

    # iterate fields and find env vars
    kwargs: Dict[str, Any] = {}
    for field in dataclasses.fields(dclass):
        if not field.init:
            continue  # don't try to get a field that can't be set via __init__
        # get value
        try:
            env_val = os.environ[field.name]
        except KeyError:
            continue

        typ, arg_typs = field.type, None

        # detect bare 'Final' and 'Optional'
        if isinstance(typ, _SpecialForm):
            raise ValueError(
                f"'{field.type}' is not a supported type: "
                f"field='{field.name}' (any of the typing-module's SpecialForm "
                f"types, 'Final' and 'Optional', must have a nested type attached)"
            )

        # take care of 'typing'-module types
        if isinstance(typ, GenericAlias):
            # Ex: Final[int], Optional[Dict[str,int]]
            if _is_optional(typ) or _is_final(typ):
                if isinstance(typ.__args__[0], type):  # Ex: Final[int], Optional[int]
                    typ, arg_typs = typ.__args__[0], None
                else:  # Final[Dict[str,int]], Optional[Dict[str,int]]
                    typ, arg_typs = typ.__args__[0].__origin__, typ.__args__[0].__args__
            # Ex: List[int], Dict[str,int]
            else:
                typ, arg_typs = typ.__origin__, typ.__args__
            if not (
                isinstance(typ, type)
                and (arg_typs is None or all(isinstance(x, type) for x in arg_typs))
            ):
                raise ValueError(
                    f"'{field.type}' is not a supported type: "
                    f"field='{field.name}' (the typing-module's alias "
                    f"types must resolve to 'type' within 1 nesting, "
                    f"or 2 if using 'Final' or 'Optional')"
                )

        try:
            kwargs[field.name] = _typecast_for_dataclass(
                env_val, typ, arg_typs, collection_sep, dict_kv_joiner
            )
        except ValueError as e:
            raise ValueError(
                f"'{field.type}'-indicated value is not a legal value: "
                f"var='{field.name}' value='{env_val}'"
            ) from e

    try:
        return dclass(**kwargs)
    except TypeError as e:
        m = re.fullmatch(
            r".*__init__\(\) missing \d+ required positional argument(?P<s>s?): (?P<args>.+)",
            str(e),
        )  # in 3.10 the class's qualname is used before "__init__()..."
        if m:
            raise OSError(
                f"Missing required environment variable{m.groupdict()['s']}: "
                f"{m.groupdict()['args']}"
            ) from e
        raise  # some other kind of TypeError



"""DEPRECATED: Module to support the `setuptools.setup` utility within `setup.py` files."""


import os
import pprint
import re
import sys
from typing import Any, Dict, List, Optional, Tuple

try:
    from typing import TypedDict
except ImportError:  # pre-3.8 support
    TypedDict = Dict[str, Any]

PythonVersion = Tuple[int, int]


class SetupShopKwargs(TypedDict):
    """The auto-created kwargs for `setuptools.setup()`."""

    name: str
    version: str
    author: str
    author_email: str
    description: str
    long_description: str
    long_description_content_type: str
    keywords: List[str]
    classifiers: List[str]
    license: str
    packages: List[str]
    install_requires: List[str]


class SetupShop:
    """Programmatically construct arguments for use in `setuptools.setup()`.

    This class is not intended to replace `setuptools.setup()`, but
    rather supplement more complex boilerplate code to reduce errors.
    All computation (IO things, compatibility checks, etc.) is done
    up-front in the constructor; so if a SetupShop instance is made,
    you're go to go.

    All required packages (`install_requires` list) are found by parsing
    `requirements.txt`. PyPI packages are assumed to be backwards-
    compatible, so these use the indicated version as a *MINIMAL*
    requirement (`==` is replaced with `>=`; this can be turned off by
    adding each package to `pinned_packages`). Conversely, GitHub-URL
    packages ARE pinned to their indicated version/tag. These packages
    are re-parsed to point to the standard `.zip` file/url.

    Example:
        `shop = SetupShop(...)`
        `setuptools.setup(..., **shop.get_kwargs(...))`

    Arguments:
        package_name -- the name of your package
        abspath_to_root -- use: `os.path.abspath(os.path.dirname(__file__))`
        py_min_max -- the min and max supported python releases: Ex: `((3,6), (3,8))`
        description -- a one-line description of your package

    Keyword arguments:
        allow_git_urls -- whether to allow github URLs in `install_requires` list
        pinned_packages -- packages to pin with version indicated in `requirements.txt`
                             (does not auto-upgrade package version)
    """

    def __init__(
        self,
        package_name: str,
        abspath_to_root: str,
        py_min_max: Tuple[PythonVersion, PythonVersion],
        description: str,
        allow_git_urls: bool = True,
        pinned_packages: Optional[List[str]] = None,
    ):
        print(  # DEPRECATION WARNING
            "\033[93m"
            + "\033[1m"
            + "SetupShop is deprecated. Use GH Action: WIPACrepo/wipac-dev-py-setup-action"
            + "\033[0m"
        )

        py_min, py_max = SetupShop._get_py_min_max(py_min_max)

        if not re.match(r"\w+$", package_name):
            raise Exception(f"Package name contains illegal characters: {package_name}")
        self.name = package_name

        # Before anything else, check that the current python version is okay
        SetupShop._ensure_python_compatibility(self.name, py_min, py_max)

        if not abspath_to_root.startswith("/"):
            raise Exception(
                f"Path is not absolute: `{abspath_to_root}`; "
                "use: `os.path.abspath(os.path.dirname(__file__))`"
            )
        self._here = abspath_to_root

        self._version = SetupShop._get_version(self._here, self.name)
        print(f"SetupShop --> version: {self._version}")

        # Make Description(s)
        self._description = description
        # include new-lines in long description
        try:
            self._readme = os.path.join(self._here, "README.md")
            self._long_description = open(self._readme).read()
        except FileNotFoundError:
            self._readme = os.path.join(self._here, "README.rst")
            self._long_description = open(self._readme).read()
        print(
            f"SetupShop --> long_description: {len(self._long_description.splitlines())} lines "
            f"(from {self._readme})"
        )

        # Gather Classifiers List
        self._classifiers = SetupShop._get_py_classifiers(py_min, py_max)
        self._classifiers.append(SetupShop._get_development_status(self._version))
        for pyc in self._classifiers:
            print(f"SetupShop --> classifier: '{pyc}'")

        # Parse requirements.txt -> 'install_requires'
        self._install_requires, req_path = SetupShop._get_install_requires(
            self._here,
            self.name,
            allow_git_urls,
            pinned_packages if pinned_packages else [],
        )
        print(
            f"SetupShop --> install_requires: {len(self._install_requires)} packages "
            f"(from {req_path})"
        )

        print()

    @staticmethod
    def _get_py_min_max(
        py_min_max: Tuple[PythonVersion, PythonVersion]
    ) -> Tuple[PythonVersion, PythonVersion]:
        """Check that the given `get_py_min_max` is valid, then return."""
        if (
            len(py_min_max) != 2
            or py_min_max[0] > py_min_max[1]
            or any(len(p) != 2 for p in py_min_max)
        ):
            raise Exception(
                "'py_min_max' must be a 2-tuple of non-decreasing 2-tuples; "
                "examples: `((3,6),(3,8))` or `((3,6),(3,6))`"
            )
        return py_min_max

    @staticmethod
    def _ensure_python_compatibility(
        name: str, py_min: PythonVersion, py_max: PythonVersion
    ) -> None:
        """If current python version is not compatible, warn and/or exit."""
        current = (sys.version_info.major, sys.version_info.minor)  # ignore micro

        if current < py_min:  # Ex: 3.5.2 < 3.5
            raise Exception(
                f"{name} requires at least Python "
                f"{py_min[0]}.{py_min[1]} to run "
                f"(current={current})"
            )
        elif current > py_max:  # ignore micro
            print(
                f"WARNING: {name} does not officially support Python "
                f"{current[0]}.{current[1]}+ "
                f"(max={py_max})"
            )

    @staticmethod
    def _find_file(here: str, pkg_name: str, fname: str) -> str:
        """Find the file `fname` and return its path."""
        # check 'here'
        if fname in os.listdir(here):
            return os.path.join(here, fname)
        # check 'here/pkg_name'
        elif fname in os.listdir(os.path.join(here, pkg_name)):
            return os.path.join(here, pkg_name, fname)
        else:
            raise FileNotFoundError(
                f"'{fname}' not found: "
                f"it can either be in '{here}' or '{os.path.join(here, pkg_name)}'"
            )

    @staticmethod
    def _get_version(here: str, name: str) -> str:
        """Get the package's `__version__` string.

        `__version__` needs to be parsed as plain text due to potential
        race condition, see:
        https://stackoverflow.com/a/2073599/13156561
        """
        init_path = SetupShop._find_file(here, name, "__init__.py")

        with open(init_path) as f:
            for line in f.readlines():
                if "__version__" in line:
                    # grab "X.Y.Z" from `__version__ = 'X.Y.Z'`
                    # - quote-style insensitive
                    return line.replace('"', "'").split("=")[-1].split("'")[1]

        raise Exception(
            f"cannot find __version__ in lowest-level __init__.py ({init_path})"
        )

    @staticmethod
    def _get_install_requires(
        here: str, name: str, allow_git_urls: bool, pinned_packages: List[str]
    ) -> Tuple[List[str], str]:
        """Get the `install_requires` list & the path to requirements.txt."""
        reqs_path = SetupShop._find_file(here, name, "requirements.txt")

        def parse_package_name(req: str) -> str:
            # https://www.python.org/dev/peps/pep-0508/#names
            rematch = re.match(r"^[A-Za-z0-9._-]+", req)
            if not rematch:
                raise Exception(f"Malformed package requirement line: '{req}'")
            return rematch.group(0)

        def convert(req: str) -> str:
            # GitHub Packages
            if "github.com" in req:
                if not allow_git_urls:
                    raise Exception(
                        "This package cannot contain any git/github url dependencies. "
                        "This is to prevent any circular dependencies. "
                        f"The culprit: {req} from {reqs_path}"
                    )
                pat = r"^git\+(?P<url>https://github\.com/[^/]+/[^/]+)@(?P<tag>(v)?\d+\.\d+\.\d+)#egg=(?P<package>\w+)$"
                re_match = re.match(pat, req)
                if not re_match:
                    raise Exception(
                        f"from {reqs_path}: "
                        f"pip-install git-package url is not in standardized format {pat} ({req})"
                    )
                groups = re_match.groupdict()
                # point right to .zip (https://stackoverflow.com/a/56635563/13156561)
                return f'{groups["package"]} @ {groups["url"]}/archive/refs/tags/{groups["tag"]}.zip'
            # PyPI Packages: my-package==5.6.7
            else:
                if parse_package_name(req) in pinned_packages:
                    return req
                return req.replace("==", ">=")

        return [convert(m) for m in open(reqs_path).read().splitlines()], reqs_path

    @staticmethod
    def _get_py_classifiers(py_min: PythonVersion, py_max: PythonVersion) -> List[str]:
        """Get auto-detected `Programming Language :: Python :: *` list.

        NOTE: Will not work after the '3.* -> 4.0'-transition.
        """
        if py_min[0] < 3:
            raise Exception("Python-classifier automation does not work for python <3.")
        if py_max[0] >= 4:
            raise Exception("Python-classifier automation does not work for python 4+.")

        return [
            f"Programming Language :: Python :: 3.{r}"
            for r in range(py_min[1], py_max[1] + 1)
        ]

    @staticmethod
    def _get_development_status(version: str) -> str:
        """Detect the development status from the package's version.

        Known Statuses (not all are supported by `SetupShop`):
            `"Development Status :: 1 - Planning"`
            `"Development Status :: 2 - Pre-Alpha"`
            `"Development Status :: 3 - Alpha"`
            `"Development Status :: 4 - Beta"`
            `"Development Status :: 5 - Production/Stable"`
            `"Development Status :: 6 - Mature"`
            `"Development Status :: 7 - Inactive"`
        """
        if version.startswith("0.0.0"):
            return "Development Status :: 2 - Pre-Alpha"
        elif version.startswith("0.0."):
            return "Development Status :: 3 - Alpha"
        elif version.startswith("0."):
            return "Development Status :: 4 - Beta"
        elif int(version.split(".")[0]) >= 1:
            return "Development Status :: 5 - Production/Stable"
        else:
            raise Exception(
                f"Could not figure Development Status for version: {version}"
            )

    @staticmethod
    def _get_packages(name: str, subpackages: Optional[List[str]]) -> List[str]:
        """Return an aggregated list of packages.

        Optionally, include the given sub-packages now fully-prefixed
        with the main package's name.
        """

        def ensure_full_prefix(sub: str) -> str:
            if sub.startswith(f"{name}."):
                return sub
            return f"{name}.{sub}"

        pkgs = [name]
        if subpackages:
            pkgs.extend(ensure_full_prefix(p) for p in subpackages)

        return pkgs

    def get_kwargs(
        self,
        other_classifiers: Optional[List[str]] = None,
        subpackages: Optional[List[str]] = None,
    ) -> SetupShopKwargs:
        """Return a dict of auto-created arguments for `setuptools.setup()`.

        Simply collate already auto-created attributes with optionally
        given keyword arguments. Apply like: `setup(..., **shop.get_kwargs())`

        NOTE: There should be no exceptions raised.
        """
        keywords = self._description.split() + self.name.split("_")

        if not other_classifiers:
            other_classifiers = []

        if self._readme.endswith(".md"):
            long_description_content_type = "text/markdown"
        elif self._readme.endswith(".rst"):
            long_description_content_type = "text/x-rst"
        else:
            long_description_content_type = "text/plain"

        kwargs: SetupShopKwargs = {
            "name": self.name,
            "version": self._version,
            "author": "IceCube Collaboration",
            "author_email": "developers@icecube.wisc.edu",
            "description": self._description,
            "long_description": self._long_description,
            "long_description_content_type": long_description_content_type,
            "keywords": keywords,
            "classifiers": sorted(
                self._classifiers
                + other_classifiers
                + ["License :: OSI Approved :: MIT License"]
            ),
            "license": "MIT",
            "packages": SetupShop._get_packages(self.name, subpackages),
            "install_requires": self._install_requires,
        }

        print("SetupShop Kwargs...")
        pprint.pprint(kwargs)
        print()
        return kwargs



"""Init."""

from . import logging_tools
from .enviro_tools import from_environment, from_environment_as_dataclass  # noqa
from .setup_tools import SetupShop  # noqa
from .strtobool import strtobool

__all__ = [
    "from_environment",
    "from_environment_as_dataclass",
    "SetupShop",
    "logging_tools",
    "strtobool",
]

# version is a human-readable version number.
__version__ = "1.4.0"

# version_info is a four-tuple for programmatic comparison. The first
# three numbers are the components of the version number. The fourth
# is zero for an official release, positive for a development branch,
# or negative for a release candidate or beta (after the base version
# number has been incremented)
version_info = (
    int(__version__.split(".")[0]),
    int(__version__.split(".")[1]),
    int(__version__.split(".")[2]),
    0,
)
